# Generated by bootstrap.sh from spead2/tools/bench_asyncio.py. Do not edit.

# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Benchmark tool to estimate the sustainable SPEAD bandwidth between two
machines, for a specific set of configurations.

Since UDP is lossy, this is not a trivial problem. We binary search for the
speed that is just sustainable. To make the test at a specific speed more
reliable, it is repeated several times, opening a new stream each time, and
with a delay to allow processors to return to idle states. A TCP control
stream is used to synchronise the two ends. All configuration is done on
the master end.
"""

from __future__ import division, print_function
import numpy as np
import spead2
import spead2.recv
import spead2.recv.asyncio
import spead2.send
import spead2.send.asyncio
import argparse
import json
import asyncio

import collections
import logging
import traceback
import timeit
import six


class SlaveConnection(object):
    def __init__(self, reader, writer):
        self.reader = reader
        self.writer = writer

    @asyncio.coroutine
    def run_stream(self, stream):
        num_heaps = 0
        while True:
            try:
                heap = yield from(stream.get())
                num_heaps += 1
            except spead2.Stopped:
                return (num_heaps)

    def _write(self, s):
        self.writer.write(s.encode('ascii'))

    @asyncio.coroutine
    def run_control(self):
        try:
            stream_task = None
            while True:
                command = yield from(self.reader.readline())
                command = command.decode('ascii')
                logging.debug("command = %s", command)
                if not command:
                    break
                command = json.loads(command)
                if command['cmd'] == 'start':
                    if stream_task is not None:
                        logging.warning("Start received while already running: %s", command)
                        continue
                    args = argparse.Namespace(**command['args'])
                    if six.PY2:
                        args_dict = vars(args)
                        for key in args_dict:
                            if isinstance(args_dict[key], unicode):
                                args_dict[key] = args_dict[key].encode('ascii')
                    if args.recv_affinity is not None and len(args.recv_affinity) > 0:
                        spead2.ThreadPool.set_affinity(args.recv_affinity[0])
                        thread_pool = spead2.ThreadPool(1, args.recv_affinity[1:] + args.recv_affinity[:1])
                    else:
                        thread_pool = spead2.ThreadPool()
                    thread_pool = spead2.ThreadPool()
                    memory_pool = spead2.MemoryPool(
                        args.heap_size, args.heap_size + 1024, args.mem_max_free, args.mem_initial)
                    stream = spead2.recv.asyncio.Stream(thread_pool, 0, args.heaps, args.ring_heaps)
                    stream.set_memory_allocator(memory_pool)
                    if args.memcpy_nt:
                        stream.set_memcpy(spead2.MEMCPY_NONTEMPORAL)
                    bind_hostname = '' if args.multicast is None else args.multicast
                    if 'recv_ibv' in args and args.recv_ibv is not None:
                        try:
                            stream.add_udp_ibv_reader(
                                bind_hostname, args.port,
                                args.recv_ibv,
                                args.packet, args.recv_buffer,
                                args.recv_ibv_vector, args.recv_ibv_max_poll)
                        except AttributeError:
                            logging.error('--recv-ibv passed but slave does not support ibv')
                            sys.exit(1)
                    else:
                        stream.add_udp_reader(args.port, args.packet, args.recv_buffer, bind_hostname)
                    thread_pool = None
                    memory_pool = None
                    stream_task = asyncio.async(self.run_stream(stream))
                    self._write('ready\n')
                elif command['cmd'] == 'stop':
                    if stream_task is None:
                        logging.warning("Stop received when already stopped")
                        continue
                    stream.stop()
                    received_heaps = yield from(stream_task)
                    self._write(json.dumps({'received_heaps': received_heaps}) + '\n')
                    stream_task = None
                    stream = None
                elif command['cmd'] == 'exit':
                    break
                else:
                    logging.warning("Bad command: %s", command)
                    continue
            logging.debug("Connection closed")
            if stream_task is not None:
                stream.stop()
                yield from(stream_task)
        except Exception:
            traceback.print_exc()


@asyncio.coroutine
def slave_connection(reader, writer):
    try:
        conn = SlaveConnection(reader, writer)
        yield from(conn.run_control())
    except Exception:
        traceback.print_exc()


def run_slave(args):
    server = yield from(asyncio.start_server(slave_connection, port=args.port))
    yield from(server.wait_closed())


@asyncio.coroutine
def send_stream(item_group, stream, num_heaps):
    tasks = collections.deque()
    transferred = 0
    for i in range(num_heaps + 1):
        while len(tasks) >= 2:
            transferred += yield from(tasks.popleft())
        if i == num_heaps:
            heap = item_group.get_end()
        else:
            heap = item_group.get_heap(data='all')
        task = asyncio.async(stream.async_send_heap(heap))
        tasks.append(task)
    for task in tasks:
        transferred += yield from(task)
    return (transferred)


def measure_connection_once(args, rate, num_heaps, required_heaps):
    reader, writer = yield from(asyncio.open_connection(args.host, args.port))
    def write(s):
        writer.write(s.encode('ascii'))
    write(json.dumps({'cmd': 'start', 'args': vars(args)}) + '\n')
    # Wait for "ready" response
    response = yield from(reader.readline())
    assert response == b'ready\n'
    if args.send_affinity is not None and len(args.send_affinity) > 0:
        spead2.ThreadPool.set_affinity(args.send_affinity[0])
        thread_pool = spead2.ThreadPool(1, args.send_affinity[1:] + args.send_affinity[:1])
    else:
        thread_pool = spead2.ThreadPool()
    thread_pool = spead2.ThreadPool()
    config = spead2.send.StreamConfig(
        max_packet_size=args.packet,
        burst_size=args.burst,
        rate=rate,
        max_heaps=num_heaps + 1)
    host = args.host
    if args.multicast is not None:
        host = args.multicast
    if 'send_ibv' in args and args.send_ibv is not None:
        stream = spead2.send.asyncio.UdpIbvStream(
            thread_pool, host, args.port, config, args.send_ibv, args.send_buffer,
            1, args.send_ibv_vector, args.send_ibv_max_poll)
    else:
        stream = spead2.send.asyncio.UdpStream(
            thread_pool, host, args.port, config, args.send_buffer)
    item_group = spead2.send.ItemGroup(
        flavour=spead2.Flavour(4, 64, args.addr_bits, 0))
    item_group.add_item(id=None, name='Test item',
                        description='A test item with arbitrary value',
                        shape=(args.heap_size,), dtype=np.uint8,
                        value=np.zeros((args.heap_size,), dtype=np.uint8))

    start = timeit.default_timer()
    transferred = yield from(send_stream(item_group, stream, num_heaps))
    end = timeit.default_timer()
    elapsed = end - start
    actual_rate = transferred / elapsed
    # Give receiver time to catch up with any queue
    yield from(asyncio.sleep(0.1))
    write(json.dumps({'cmd': 'stop'}) + '\n')
    # Read number of heaps received
    response = yield from(reader.readline())
    response = json.loads(response.decode('ascii'))
    received_heaps = response['received_heaps']
    yield from(asyncio.sleep(0.5))
    yield from(writer.drain())
    writer.close()
    logging.debug("Received %d/%d heaps in %f seconds, rate %.3f Gbps",
            received_heaps, num_heaps, elapsed, actual_rate * 8e-9)
    return (received_heaps >= required_heaps, actual_rate)


def measure_connection(args, rate, num_heaps, required_heaps):
    good = True
    rate_sum = 0.0
    passes = 5
    for i in range(passes):
        status, actual_rate = yield from(measure_connection_once(args, rate, num_heaps, required_heaps))
        good = good and status
        rate_sum += actual_rate
    return (good, rate_sum / passes)


def run_master(args):
    best_actual = 0.0

    # Send 1GB as fast as possible to find an upper bound - receive rate
    # does not matter. Also do a warmup run first to warm up the receiver.
    num_heaps = int(1e9 / args.heap_size) + 2
    yield from(measure_connection_once(args, 0.0, num_heaps, 0))  # warmup
    good, actual_rate = yield from(measure_connection(args, 0.0, num_heaps, num_heaps - 1))
    if good:
        if not args.quiet:
            print("Limited by send spead")
        best_actual = actual_rate
    else:
        print("Send rate: {:.3f} Gbps".format(actual_rate * 8e-9))
        low = 0.0
        high = actual_rate
        while high - low > high * 0.02:
            # Need at least 1GB of data to overwhelm cache effects, and want at least
            # 1 second for warmup effects.
            rate = (low + high) * 0.5
            num_heaps = int(max(1e9, rate) / args.heap_size) + 2
            good, actual_rate = yield from(measure_connection(args, rate, num_heaps, num_heaps - 1))
            if not args.quiet:
                print("Rate: {:.3f} Gbps ({:.3f} actual): {}".format(
                    rate * 8e-9, actual_rate * 8e-9, "GOOD" if good else "BAD"))
            if good:
                low = rate
                best_actual = actual_rate
            else:
                high = rate
    rate_gbps = best_actual * 8e-9
    if args.quiet:
        print(rate_gbps)
    else:
        print("Sustainable rate: {:.3f} Gbps".format(rate_gbps))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--log', metavar='LEVEL', default='INFO', help='Log level [%(default)s]')
    subparsers = parser.add_subparsers(title='subcommands')
    master = subparsers.add_parser('master')
    master.add_argument('--quiet', action='store_true', default=False, help='Print only the final result')
    master.add_argument('--packet', metavar='BYTES', type=int, default=9172, help='Maximum packet size to use for UDP [%(default)s]')
    master.add_argument('--heap-size', metavar='BYTES', type=int, default=4194304, help='Payload size for heap [%(default)s]')
    master.add_argument('--addr-bits', metavar='BITS', type=int, default=40, help='Heap address bits [%(default)s]')
    master.add_argument('--multicast', metavar='ADDRESS', type=str, help='Send via multicast group [unicast]')
    group = master.add_argument_group('sender options')
    group.add_argument('--send-affinity', type=spead2.parse_range_list, help='List of CPUs to pin threads to [no affinity]')
    group.add_argument('--send-buffer', metavar='BYTES', type=int, default=spead2.send.asyncio.UdpStream.DEFAULT_BUFFER_SIZE, help='Socket buffer size [%(default)s]')
    group.add_argument('--burst', metavar='BYTES', type=int, default=spead2.send.StreamConfig.DEFAULT_BURST_SIZE, help='Send burst size [%(default)s]')
    if hasattr(spead2.send, 'UdpIbvStream'):
        group.add_argument('--send-ibv', type=str, metavar='ADDRESS', help='Use ibverbs with this interface address [no]')
        group.add_argument('--send-ibv-vector', type=int, default=0, metavar='N', help='Completion vector, or -1 to use polling [%(default)s]')
        group.add_argument('--send-ibv-max-poll', type=int, default=spead2.send.UdpIbvStream.DEFAULT_MAX_POLL, help='Maximum number of times to poll in a row [%(default)s]')
    group = master.add_argument_group('receiver options')
    group.add_argument('--recv-affinity', type=spead2.parse_range_list, help='List of CPUs to pin threads to [no affinity]')
    group.add_argument('--recv-buffer', metavar='BYTES', type=int, default=spead2.recv.Stream.DEFAULT_UDP_BUFFER_SIZE, help='Socket buffer size [%(default)s]')
    if hasattr(spead2.recv.Stream, 'add_udp_ibv_reader'):
        group.add_argument('--recv-ibv', type=str, metavar='ADDRESS', help='Use ibverbs with this interface address [no]')
        group.add_argument('--recv-ibv-vector', type=int, default=0, metavar='N', help='Completion vector, or -1 to use polling [%(default)s]')
        group.add_argument('--recv-ibv-max-poll', type=int, default=spead2.recv.Stream.DEFAULT_UDP_IBV_MAX_POLL, help='Maximum number of times to poll in a row [%(default)s]')
    group.add_argument('--heaps', type=int, default=spead2.recv.Stream.DEFAULT_MAX_HEAPS, help='Maximum number of in-flight heaps [%(default)s]')
    group.add_argument('--ring-heaps', type=int, default=spead2.recv.Stream.DEFAULT_RING_HEAPS, help='Ring buffer capacity in heaps [%(default)s]')
    group.add_argument('--memcpy-nt', action='store_true', help='Use non-temporal memcpy [no]')
    group.add_argument('--mem-max-free', type=int, default=12, help='Maximum free memory buffers [%(default)s]')
    group.add_argument('--mem-initial', type=int, default=8, help='Initial free memory buffers [%(default)s]')
    master.add_argument('host')
    master.add_argument('port', type=int)
    slave = subparsers.add_parser('slave')
    slave.add_argument('port', type=int)

    args = parser.parse_args()
    logging.basicConfig(level=getattr(logging, args.log.upper()))
    if 'host' in args:
        task = run_master(args)
    else:
        task = run_slave(args)
    task = asyncio.async(task)
    asyncio.get_event_loop().run_until_complete(task)
    task.result()



# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for parts of spead2 that are shared between send and receive"""

from __future__ import division, print_function
import spead2
import numpy as np
import six
from nose.tools import *


def assert_equal_typed(expected, actual, msg=None):
    """Check that expected and actual compare equal *and* have the same type.

    This is used for checking that strings have the correct type (str vs
    unicode in Python 2, str vs bytes in Python 3).
    """
    assert_equal(expected, actual, msg)
    assert_equal(type(expected), type(actual), msg)


class TestParseRangeList(object):
    def test_empty(self):
        assert_equal([], spead2.parse_range_list(''))

    def test_simple(self):
        assert_equal([1, 2, 5], spead2.parse_range_list('1,2,5'))

    def test_ranges(self):
        assert_equal([100, 4, 5, 6, 8, 10, 12, 13], spead2.parse_range_list('100,4-6,8,10-10,12-13'))


class TestThreadPool(object):
    """Smoke tests for :py:class:`spead2.ThreadPool`. These are very simple
    tests, because it is not actually possible to check things like the
    thread affinity."""
    def test_simple(self):
        spead2.ThreadPool()
        spead2.ThreadPool(4)

    def test_affinity(self):
        spead2.ThreadPool(3, [0, 1])
        spead2.ThreadPool(1, [1, 0, 2])


class TestFlavour(object):
    def test_bad_version(self):
        with assert_raises(ValueError):
            spead2.Flavour(3, 64, 40, 0)

    def test_bad_item_pointer_bits(self):
        with assert_raises(ValueError):
            spead2.Flavour(4, 32, 24, 0)

    def test_bad_heap_address_bits(self):
        with assert_raises(ValueError):
            spead2.Flavour(4, 64, 43, 0)  # Not multiple of 8
        with assert_raises(ValueError):
            spead2.Flavour(4, 64, 0, 0)
        with assert_raises(ValueError):
            spead2.Flavour(4, 64, 64, 0)

    def test_attributes(self):
        flavour = spead2.Flavour(4, 64, 40, 2)
        assert_equal(4, flavour.version)
        assert_equal(64, flavour.item_pointer_bits)
        assert_equal(40, flavour.heap_address_bits)
        assert_equal(2, flavour.bug_compat)

    def test_equality(self):
        flavour1 = spead2.Flavour(4, 64, 40, 2)
        flavour1b = spead2.Flavour(4, 64, 40, 2)
        flavour2 = spead2.Flavour(4, 64, 48, 2)
        flavour3 = spead2.Flavour(4, 64, 40, 4)
        assert_true(flavour1 == flavour1b)
        assert_true(flavour1 != flavour2)
        assert_false(flavour1 != flavour1b)
        assert_false(flavour1 == flavour2)
        assert_false(flavour1 == flavour3)


class TestItem(object):
    """Tests for :py:class:`spead2.Item`.

    Many of these actually test :py:class:`spead2.Descriptor`, but since the
    internals of these classes are interwined, it is simpler to keep all the
    tests together here."""

    def test_nonascii_value(self):
        """Using a non-ASCII unicode character raises a
        :py:exc:`UnicodeEncodeError`."""
        item1 = spead2.Item(0x1000, 'name1', 'description',
            (None,), format=[('c', 8)], value=six.u('\u0200'))
        item2 = spead2.Item(0x1001, 'name2', 'description2', (),
            dtype='S5', value=six.u('\u0201'))
        assert_raises(UnicodeEncodeError, item1.to_buffer)
        assert_raises(UnicodeEncodeError, item2.to_buffer)

    def test_format_and_dtype(self):
        """Specifying both a format and dtype raises :py:exc:`ValueError`."""
        assert_raises(ValueError, spead2.Item, 0x1000, 'name', 'description',
            (1, 2), format=[('c', 8)], dtype='S1')

    def test_no_format_or_dtype(self):
        """At least one of format and dtype must be specified."""
        assert_raises(ValueError, spead2.Item, 0x1000, 'name', 'description',
            (1, 2), format=None)

    def test_invalid_order(self):
        """The `order` parameter must be either 'C' or 'F'."""
        assert_raises(ValueError, spead2.Item, 0x1000, 'name', 'description',
            (1, 2), np.int32, order='K')

    def test_fortran_fallback(self):
        """The `order` parameter must be either 'C' for legacy formats."""
        assert_raises(ValueError, spead2.Item, 0x1000, 'name', 'description',
            (1, 2), format=[('u', 32)], order='F')

    def test_empty_format(self):
        """Format must not be empty"""
        assert_raises(ValueError, spead2.Item, 0x1000, 'name', 'description',
            (1, 2), format=[])

    def test_assign_none(self):
        """Changing a value back to `None` raises :py:exc:`ValueError`."""
        item = spead2.Item(0x1000, 'name', 'description', (), np.int32)
        with assert_raises(ValueError):
            item.value = None

    def test_multiple_unknown(self):
        """Multiple unknown dimensions are not allowed."""
        assert_raises(ValueError, spead2.Item, 0x1000, 'name', 'description',
            (5, None, 3, None), format=[('u', 32)])

    def test_numpy_unknown(self):
        """Unknown dimensions are not permitted when using a numpy descriptor"""
        assert_raises(ValueError, spead2.Item, 0x1000, 'name', 'description',
            (5, None), np.int32)

    def test_nonascii_name(self):
        """Name with non-ASCII characters must fail"""
        with assert_raises(UnicodeEncodeError):
            item = spead2.Item(0x1000, six.u('\u0200'), 'description', (), np.int32)
            item.to_raw(spead2.Flavour())

    def test_nonascii_description(self):
        """Description with non-ASCII characters must fail"""
        with assert_raises(UnicodeEncodeError):
            item = spead2.Item(0x1000, 'name', six.u('\u0200'), (), np.int32)
            item.to_raw(spead2.Flavour())


class TestItemGroup(object):
    """Tests for :py:class:`spead2.ItemGroup`"""

    def test_allocate_id(self):
        """Automatic allocation of IDs must skip over already-allocated IDs"""
        ig = spead2.ItemGroup()
        ig.add_item(0x1000, 'item 1', 'item 1', (), np.int32)
        ig.add_item(0x1003, 'item 2', 'item 2', (), np.int32)
        ig.add_item(None, 'item 3', 'item 3', (), np.int32)
        ig.add_item(None, 'item 4', 'item 4', (), np.int32)
        ig.add_item(None, 'item 5', 'item 5', (), np.int32)
        assert_equal('item 3', ig[0x1001].name)
        assert_equal('item 4', ig[0x1002].name)
        assert_equal('item 5', ig[0x1004].name)

    def test_replace_rename(self):
        """When a new item is added with a known ID but a different name, the
        old item must cease to exist."""
        ig = spead2.ItemGroup()
        ig.add_item(0x1000, 'item 1', 'item 1', (), np.int32)
        ig.add_item(0x1000, 'renamed', 'renamed', (), np.int32)
        assert_equal([0x1000], list(ig.ids()))
        assert_equal(['renamed'], list(ig.keys()))

    def test_replace_clobber_name(self):
        """When a new item is added that collides with an existing name, that
        other item is deleted."""
        ig = spead2.ItemGroup()
        ig.add_item(0x1000, 'item 1', 'item 1', (), np.int32)
        ig.add_item(0x1001, 'item 1', 'clobbered', (), np.int32)
        assert_equal([0x1001], list(ig.ids()))
        assert_equal(['item 1'], list(ig.keys()))
        assert_equal('clobbered', ig['item 1'].description)
        assert_is(ig['item 1'], ig[0x1001])

    def test_replace_reset_value(self):
        """When a descriptor is replaced with an identical one but a new
        value, the new value must take effect and the version must be
        incremented."""
        ig = spead2.ItemGroup()
        ig.add_item(0x1000, 'item 1', 'item 1', (), np.int32, value=np.int32(4))
        ig.add_item(0x1001, 'item 2', 'item 2', (), np.int32, value=np.int32(5))
        item1 = ig[0x1000]
        item2 = ig[0x1001]
        item1_version = item1.version
        item2_version = item2.version
        ig.add_item(0x1000, 'item 1', 'item 1', (), np.int32, value=np.int32(6))
        ig.add_item(0x1001, 'item 2', 'item 2', (), np.int32)
        assert_is(item1, ig[0x1000])
        assert_is(item2, ig[0x1001])
        assert_equal(np.int32(6), item1.value)
        assert_greater(item1.version, item1_version)
        assert_equal(np.int32(5), item2.value)
        assert_equal(item2.version, item2_version)

    def test_replace_change_shape(self):
        """When a descriptor changes the shape, the old item must be discarded
        and ``None`` used in its place. The version must be bumped."""
        ig = spead2.ItemGroup()
        ig.add_item(0x1000, 'item 1', 'item 1', (), np.int32, value=np.int32(4))
        item1 = ig[0x1000]
        item1_version = item1.version
        ig.add_item(0x1000, 'item 1', 'bigger', (3, 4), np.int32)
        assert_is_not(item1, ig[0x1000])
        assert_is_none(ig[0x1000].value)
        assert_greater(ig[0x1000].version, item1_version)

    def test_replace_clobber_both(self):
        """Adding a new item that collides with one item on name and another on
        ID causes both to be dropped."""
        ig = spead2.ItemGroup()
        ig.add_item(0x1000, 'item 1', 'item 1', (), np.int32)
        ig.add_item(0x1001, 'item 2', 'item 2', (), np.int32)
        ig.add_item(0x1000, 'item 2', 'clobber', (), np.int32)
        assert_equal([0x1000], list(ig.ids()))
        assert_equal(['item 2'], list(ig.keys()))
        assert_is(ig[0x1000], ig['item 2'])
        assert_equal('clobber', ig[0x1000].description)



import pytest
from pybind11_tests import kwargs_and_defaults as m


def test_function_signatures(doc):
    assert doc(m.kw_func0) == "kw_func0(arg0: int, arg1: int) -> str"
    assert doc(m.kw_func1) == "kw_func1(x: int, y: int) -> str"
    assert doc(m.kw_func2) == "kw_func2(x: int=100, y: int=200) -> str"
    assert doc(m.kw_func3) == "kw_func3(data: str='Hello world!') -> None"
    assert doc(m.kw_func4) == "kw_func4(myList: List[int]=[13, 17]) -> str"
    assert doc(m.kw_func_udl) == "kw_func_udl(x: int, y: int=300) -> str"
    assert doc(m.kw_func_udl_z) == "kw_func_udl_z(x: int, y: int=0) -> str"
    assert doc(m.args_function) == "args_function(*args) -> tuple"
    assert doc(m.args_kwargs_function) == "args_kwargs_function(*args, **kwargs) -> tuple"
    assert doc(m.KWClass.foo0) == \
        "foo0(self: m.kwargs_and_defaults.KWClass, arg0: int, arg1: float) -> None"
    assert doc(m.KWClass.foo1) == \
        "foo1(self: m.kwargs_and_defaults.KWClass, x: int, y: float) -> None"


def test_named_arguments(msg):
    assert m.kw_func0(5, 10) == "x=5, y=10"

    assert m.kw_func1(5, 10) == "x=5, y=10"
    assert m.kw_func1(5, y=10) == "x=5, y=10"
    assert m.kw_func1(y=10, x=5) == "x=5, y=10"

    assert m.kw_func2() == "x=100, y=200"
    assert m.kw_func2(5) == "x=5, y=200"
    assert m.kw_func2(x=5) == "x=5, y=200"
    assert m.kw_func2(y=10) == "x=100, y=10"
    assert m.kw_func2(5, 10) == "x=5, y=10"
    assert m.kw_func2(x=5, y=10) == "x=5, y=10"

    with pytest.raises(TypeError) as excinfo:
        # noinspection PyArgumentList
        m.kw_func2(x=5, y=10, z=12)
    assert excinfo.match(
        r'(?s)^kw_func2\(\): incompatible.*Invoked with: kwargs: ((x=5|y=10|z=12)(, |$))' + '{3}$')

    assert m.kw_func4() == "{13 17}"
    assert m.kw_func4(myList=[1, 2, 3]) == "{1 2 3}"

    assert m.kw_func_udl(x=5, y=10) == "x=5, y=10"
    assert m.kw_func_udl_z(x=5) == "x=5, y=0"


def test_arg_and_kwargs():
    args = 'arg1_value', 'arg2_value', 3
    assert m.args_function(*args) == args

    args = 'a1', 'a2'
    kwargs = dict(arg3='a3', arg4=4)
    assert m.args_kwargs_function(*args, **kwargs) == (args, kwargs)


def test_mixed_args_and_kwargs(msg):
    mpa = m.mixed_plus_args
    mpk = m.mixed_plus_kwargs
    mpak = m.mixed_plus_args_kwargs
    mpakd = m.mixed_plus_args_kwargs_defaults

    assert mpa(1, 2.5, 4, 99.5, None) == (1, 2.5, (4, 99.5, None))
    assert mpa(1, 2.5) == (1, 2.5, ())
    with pytest.raises(TypeError) as excinfo:
        assert mpa(1)
    assert msg(excinfo.value) == """
        mixed_plus_args(): incompatible function arguments. The following argument types are supported:
            1. (arg0: int, arg1: float, *args) -> tuple

        Invoked with: 1
    """  # noqa: E501 line too long
    with pytest.raises(TypeError) as excinfo:
        assert mpa()
    assert msg(excinfo.value) == """
        mixed_plus_args(): incompatible function arguments. The following argument types are supported:
            1. (arg0: int, arg1: float, *args) -> tuple

        Invoked with:
    """  # noqa: E501 line too long

    assert mpk(-2, 3.5, pi=3.14159, e=2.71828) == (-2, 3.5, {'e': 2.71828, 'pi': 3.14159})
    assert mpak(7, 7.7, 7.77, 7.777, 7.7777, minusseven=-7) == (
        7, 7.7, (7.77, 7.777, 7.7777), {'minusseven': -7})
    assert mpakd() == (1, 3.14159, (), {})
    assert mpakd(3) == (3, 3.14159, (), {})
    assert mpakd(j=2.71828) == (1, 2.71828, (), {})
    assert mpakd(k=42) == (1, 3.14159, (), {'k': 42})
    assert mpakd(1, 1, 2, 3, 5, 8, then=13, followedby=21) == (
        1, 1, (2, 3, 5, 8), {'then': 13, 'followedby': 21})
    # Arguments specified both positionally and via kwargs should fail:
    with pytest.raises(TypeError) as excinfo:
        assert mpakd(1, i=1)
    assert msg(excinfo.value) == """
        mixed_plus_args_kwargs_defaults(): incompatible function arguments. The following argument types are supported:
            1. (i: int=1, j: float=3.14159, *args, **kwargs) -> tuple

        Invoked with: 1; kwargs: i=1
    """  # noqa: E501 line too long
    with pytest.raises(TypeError) as excinfo:
        assert mpakd(1, 2, j=1)
    assert msg(excinfo.value) == """
        mixed_plus_args_kwargs_defaults(): incompatible function arguments. The following argument types are supported:
            1. (i: int=1, j: float=3.14159, *args, **kwargs) -> tuple

        Invoked with: 1, 2; kwargs: j=1
    """  # noqa: E501 line too long



from ._version import version_info, __version__  # noqa: F401 imported but unused


def get_include(*args, **kwargs):
    import os
    try:
        from pip import locations
        return os.path.dirname(
            locations.distutils_scheme('pybind11', *args, **kwargs)['headers'])
    except ImportError:
        return 'include'



# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generate and send SPEAD packets. This is mainly a benchmark application, but also
demonstrates the API."""

from __future__ import print_function, division
import spead2
import spead2.send
import spead2.send.trollius
import numpy as np
import logging
import sys
import itertools
import argparse
import trollius
import collections
from trollius import From


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('host')
    parser.add_argument('port', type=int)

    group = parser.add_argument_group('Data options')
    group.add_argument('--heap-size', metavar='BYTES', type=int, default=4194304, help='Payload size for heap [%(default)s]')
    group.add_argument('--items', type=int, default=1, help='Number of items per heap [%(default)s]')
    group.add_argument('--dtype', type=str, default='<c8', help='Numpy data type [%(default)s]')
    group.add_argument('--heaps', type=int, help='Number of data heaps to send [infinite]')

    group = parser.add_argument_group('Output options')
    group.add_argument('--log', metavar='LEVEL', default='INFO', help='Log level [%(default)s]')

    group = parser.add_argument_group('Protocol options')
    group.add_argument('--pyspead', action='store_true', help='Be bug-compatible with PySPEAD')
    group.add_argument('--addr-bits', type=int, default=40, help='Heap address bits [%(default)s]')
    group.add_argument('--packet', type=int, default=spead2.send.StreamConfig.DEFAULT_MAX_PACKET_SIZE, help='Maximum packet size to send [%(default)s]')
    group.add_argument('--descriptors', type=int, help='Description issue frequency [only at start]')

    group = parser.add_argument_group('Performance options')
    group.add_argument('--buffer', type=int, default=spead2.send.trollius.UdpStream.DEFAULT_BUFFER_SIZE, help='Socket buffer size  [%(default)s]')
    group.add_argument('--threads', type=int, default=1, help='Number of worker threads [%(default)s]')
    group.add_argument('--burst', metavar='BYTES', type=int, default=spead2.send.StreamConfig.DEFAULT_BURST_SIZE, help='Burst size [%(default)s]')
    group.add_argument('--rate', metavar='Gb/s', type=float, default=0, help='Transmission rate bound [no limit]')
    group.add_argument('--affinity', type=spead2.parse_range_list, help='List of CPUs to pin threads to [no affinity]')
    if hasattr(spead2.send, 'UdpIbvStream'):
        group.add_argument('--ibv', type=str, metavar='ADDRESS', help='Use ibverbs with this interface address [no]')
        group.add_argument('--ibv-vector', type=int, default=0, metavar='N', help='Completion vector, or -1 to use polling [%(default)s]')
        group.add_argument('--ibv-max-poll', type=int, default=spead2.send.UdpIbvStream.DEFAULT_MAX_POLL, help='Maximum number of times to poll in a row [%(default)s]')

    return parser.parse_args()


@trollius.coroutine
def run(item_group, stream, args):
    tasks = collections.deque()
    if args.heaps is None:
        rep = itertools.repeat(False)
    else:
        rep = itertools.chain(itertools.repeat(False, args.heaps), [True])
    for is_end in rep:
        if len(tasks) >= 2:
            yield From(trollius.wait([tasks.popleft()]))
        if is_end:
            task = trollius.async(stream.async_send_heap(item_group.get_end()))
        else:
            for item in item_group.values():
                item.version += 1
            task = trollius.async(stream.async_send_heap(item_group.get_heap()))
        tasks.append(task)
    while len(tasks) > 0:
        yield From(trollius.wait([tasks.popleft()]))


def main():
    args = get_args()
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    dtype = np.dtype(args.dtype)
    elements = args.heap_size // (args.items * dtype.itemsize)
    heap_size = elements * args.items * dtype.itemsize
    if heap_size != args.heap_size:
        logging.warn('Heap size is not an exact multiple: using %d instead of %d',
                     heap_size, args.heap_size)
    bug_compat = spead2.BUG_COMPAT_PYSPEAD_0_5_2 if args.pyspead else 0
    item_group = spead2.send.ItemGroup(
        descriptor_frequency=args.descriptors,
        flavour=spead2.Flavour(4, 64, args.addr_bits, bug_compat))
    for i in range(args.items):
        item_group.add_item(id=None, name='Test item {}'.format(i),
                            description='A test item with arbitrary value',
                            shape=(elements,), dtype=dtype,
                            value=np.zeros((elements,), dtype=dtype))
    if args.affinity is not None and len(args.affinity) > 0:
        spead2.ThreadPool.set_affinity(args.affinity[0])
        thread_pool = spead2.ThreadPool(args.threads, args.affinity[1:] + args.affinity[:1])
    else:
        thread_pool = spead2.ThreadPool(args.threads)
    config = spead2.send.StreamConfig(
        max_packet_size=args.packet,
        burst_size=args.burst,
        rate=args.rate * 1024**3 / 8)
    if 'ibv' in args and args.ibv is not None:
        stream = spead2.send.trollius.UdpIbvStream(
            thread_pool, args.host, args.port, config, args.ibv,
            args.buffer, args.ibv_vector, args.ibv_max_poll)
    else:
        stream = spead2.send.trollius.UdpStream(
            thread_pool, args.host, args.port, config, args.buffer)

    try:
        trollius.get_event_loop().run_until_complete(run(item_group, stream, args))
    except KeyboardInterrupt:
        sys.exit(1)
